from .exceptions import LanguageError, WordlistError
from .BIP39_List import BIP39_List
from .enums import Language
from .encode import Encode
from .decode import Decode
from hashlib import sha256
from typing import List
import os

wordlist = BIP39_List()

class Mnemonic:
    """
    Mnemonic class for holding a 24-word BIP44 seed phrase. The Mnemonic holds
    33 bytes. The first 32 bytes are the seed. The last byte is a checksum.
    """
    def __init__(self) -> None:
        """
        Initializes a new instance of the Mnemonic class.
        """
        self.seed = bytes(32)
        self.checksum = bytes(1)


    @classmethod
    def generate_random(cls) -> "Mnemonic":
        """
        Generates and returns a random mnemonic phrase using os.urandom(), which
        is suitable for cryptographic use.
        """
        mnemonic = cls()
        mnemonic.seed = os.urandom(32)
        mnemonic.checksum = sha256(mnemonic.seed).digest()[:1]

        return mnemonic


    def set_word(
            self, phrase_index: int, word: str, language: Language
            ) -> None:
        """
        Sets a word in this Mnemonic class instance according to the zero-based
        array and word given. Raises errors if the index is out of range or the
        language is not in the current list of languages.
        """
        if not isinstance(phrase_index, int):
            raise TypeError("The index argument given is not an int.")

        if phrase_index < 0 or phrase_index > 23:
            raise IndexError("The index argument given is out of bounds.")

        if not isinstance(language, Language):
            raise TypeError("The language argument was not a Language enum.")

        # Create a deleting bitmask based on the word's position in the phrase.
        max_words = 24
        word_position = phrase_index + 1
        word_count_right_side = max_words - word_position

        word_delete_bitmask = 2 ** 264 - 1
        word_delete_bitmask ^= 0b1111_1111_111 << (word_count_right_side * 11)

        # Delete the current bits at the given word index.
        mnemonic_int = Decode.mnemonic_int(self.seed, self.checksum)
        mnemonic_int &= word_delete_bitmask

        # Set the word bits at the given word index.
        word_bits = wordlist.get_word_index(word, language)
        mnemonic_int |= word_bits << (word_count_right_side * 11)

        # Recalculate the checksum if the last word, which contains the
        # checksum, is not the word being set.
        if phrase_index != 23:
            # Get the hash of the first 32 bytes.
            seed_int = mnemonic_int >> 8
            seed_bin = seed_int.to_bytes(32, "big")
            checksum_bin = sha256(seed_bin).digest()[:1]
            checksum_int = int.from_bytes(checksum_bin, "big")
            
            # Delete last byte and replace with the checksum.
            mnemonic_int >>= 8
            mnemonic_int <<= 8
            mnemonic_int &= checksum_int
            
        # Update class instance variables
        self.seed = Encode.mnemonic_seed(mnemonic_int)
        self.checksum = Encode.mnemonic_checksum(mnemonic_int)


    def get_word(self, index: int, language: Language) -> str:
        """
        Returns the word at the given zero-based index of this Mnemonic class
        instance. Raises error if the index is not between 0 and 23.
        """
        if not isinstance(language, Language):
            raise ValueError(f"{language} is not in the language list.")
            
        if not isinstance(index, int):
            raise TypeError("The index argument given is not an int.")

        if index < 0 or index > 23:
            raise IndexError("The index argument given is out of bounds.")

        # Determine number of bits to truncate from the right based the word's
        # position in the phrase.
        max_words = 24
        word_position = index + 1
        word_count_right_side = max_words - word_position

        mnemonic_int = Decode.mnemonic_int(self.seed, self.checksum)
        truncated_mnemonic_int = mnemonic_int >> (word_count_right_side * 11)

        # Get word index and return word text.
        word_bitmask = 0b1111_1111_111
        word_index = truncated_mnemonic_int & word_bitmask
        word = wordlist.get_word(word_index, language)

        return word


    def validate_phrase(phrase: List[str], language: Language) -> bool:
        """
        Returns true if the given mnemonic phrase has words that exist in the
        given language's word list, and the checksum is valid. Raises error if
        the language given is not in the current language list.
        """
        if not isinstance(language, Language):
            raise ValueError(f"{language} is not in the language list.")

        if len(phrase) != 24:
            raise ValueError("Phrase does not have 24 words")

        for word in phrase:
            if language not in wordlist.get_language(word):
                raise WordlistError(word, language)

        # Add words from left to right, shifting the added words to the left by
        # 11 bits each iteration.
        indices = [wordlist.get_word_index(word, language) for word in phrase]
        mnemonic_int = 0

        for word_index in indices:
            mnemonic_int <<= 11
            mnemonic_int += word_index

        # Validate checksum.
        given_phrase_seed = Encode.mnemonic_seed(mnemonic_int)
        given_phrase_checksum = Encode.mnemonic_checksum(mnemonic_int)
        recalculated_checksum = sha256(given_phrase_seed).digest()[:1]
        
        is_valid_checksum = given_phrase_checksum == recalculated_checksum

        return is_valid_checksum



import os
import lagrange
from typing import List
from .exceptions import ChecksumError, ThresholdError
from .enums import Checksum
from .point import Point
from .share import Share
from .encode import Encode
from .decode import Decode
from .lagrange import Lagrange
from .mnemonic import Mnemonic
from .polynomial import Polynomial


# The field size should be a prime number that is larger than the max value of
# the secret key (256 bits).
PRIME_MODULUS = 2 ** 256 - 2 ** 32 - 977


def create_shares(
        threshold: int, sharecount: int, mnemonic: Mnemonic
        ) -> List[Share]:
    """
    Splits a 24-word BIP39 mnemonic into a (k, n) threshold scheme, based on the
    Shamir Secret Sharing (SSS) system. The secret key can be recovered with any
    combination of k number of shares, but no information is revealed about the
    secret key, even with k - 1 shares.
    """
    if not isinstance(threshold, int):
        raise TypeError("The threshold argument was not of the int type.")

    if threshold < 2 or threshold > 17:
        raise ValueError("The given index argument is out of bounds.")

    if not isinstance(sharecount, int):
        raise TypeError("The sharecount argument was not of the int type.")

    if sharecount < threshold or sharecount > 256:
        raise ValueError("The given sharecount argument is out of bounds.")

    # f(x=0) is the key value.
    key_num = Decode.mnemonic_key(mnemonic.seed)
    key_point = Point(0, key_num)
    
    # f(x=1) is the key hash.
    key_hash = Encode.mnemonic_hash(key_num)
    hash_num = int.from_bytes(key_hash, "big")
    hash_point = Point(1, hash_num)

    # All f(x>1) are the share values.
    shares = []

    # Generate 2 fewer shares with random X-values than the threshold, because
    # the key value at x=0 and the hash at x=1 are already determined.
    random_share_count = threshold - 2

    for i in range(random_share_count):
        # The X-value for random shares start at x=2.
        x_val = i + 2
        random_val = int.from_bytes(os.urandom(32), "big")

        point = Point(x_val, random_val)
        share = Share(point, threshold, mnemonic.checksum)
        shares.append(share)

    # Calculate the remaining shares using Lagrange interpolation.
    base_points = []
    base_points.append(key_point)
    base_points.append(hash_point)
    base_points.extend([share.point for share in shares])
    
    calculated_share_count = sharecount - random_share_count

    for i in range(calculated_share_count):
        x_val = i + random_share_count + 2
        y_val = Lagrange.interpolate(base_points, PRIME_MODULUS, x_val)
        
        point = Point(x_val, y_val)
        share = Share(point, threshold, mnemonic.checksum)
        shares.append(share)

    return shares


def recover_key(shares: List[Point]) -> bytes:
    """
    Takes a list of Point objects, and uses Lagrange interpolation to find the
    Y-intercept, which is the secret key. If incorrect or too few Point objects
    are provided, an incorrect result will be returned. There is no error
    checking.
    """
    point_list = [(share.X, share.Y) for share in shares]
    key_int = lagrange.interpolate(point_list, PRIME_MODULUS)
    key_bin = key_int.to_bytes(32, "big")

    return key_bin


def recover_mnemonic(shares: List[Share]) -> Mnemonic:
    """
    Calculates and returns a Mnemonic object based on the given Share objects.
    Raises an error if the shares are not valid.
    """
    if not isinstance(shares, list):
        raise TypeError("The given shares argument was not a list object.")

    first_seed_checksum = shares[0].seed_checksum

    for share in shares:
        if not isinstance(share, Share):
            message = "The shares argument was not of the type List[Share]."
            raise TypeError(message)
        
        if share.seed_checksum != first_seed_checksum:
            raise ChecksumError(
                Checksum.ShareGroup, first_seed_checksum, share.seed_checksum
            )

    if len(shares) < shares[0].threshold:
        raise ThresholdError(shares[0].threshold, len(shares))

    # Recover mnemonic seed with Lagrange interpolation.
    share_points = [share.point for share in shares]

    orignal_key = Lagrange.interpolate(share_points, PRIME_MODULUS, 0)
    original_hash_int = Lagrange.interpolate(share_points, PRIME_MODULUS, 1)
    original_hash = original_hash_int.to_bytes(32, "big")
    recalculated_hash = Encode.mnemonic_hash(orignal_key)

    if original_hash != recalculated_hash:
        raise ChecksumError(
            Checksum.KeyValue, original_hash, recalculated_hash
        )

    mnemonic = Mnemonic()
    mnemonic.seed = orignal_key
    mnemonic.checksum = original_hash[:1]


from hashlib import sha256

class Encode:
    @staticmethod
    def share_X(unencoded_X: int) -> int:
        """
        Returns the given X-value as the encoded value in a Share object.
        """
        # The actual X-value starts at 2, so the encoded version is 2 less than
        # the given value.
        return unencoded_X - 2


    @staticmethod
    def share_threshold(unencoded_threshold: int) -> int:
        """
        Returns the given threshold value as the encoded value in a Share
        object.
        """
        # The actual threshold starts at 2, so the encoded version is 2 less
        # than the given value.
        encoded_threshold = unencoded_threshold - 2

        # The threshold is the second item encoded in a 2-byte sequence. The
        # X-value to its right takes 7 bits, so the threshold needs to be
        # shifted to the left by 7 bits.
        encoded_threshold <<= 7

        return encoded_threshold


    @staticmethod
    def share_version(unencoded_version: int) -> int:
        """
        Returns the given version number as the encoded value in a Share
        object.
        """
        # The version is the first item encoded in a 2-byte sequence. The
        # version and X-value to its right take 11 bits, so the version needs
        # to be shifted to the left by 11 bits.
        return unencoded_version << 11


    @staticmethod
    def share_bytes(
            encoded_X: int, encoded_Y: int, encoded_threshold: int,
            seed_checksum: bytes, encoded_version: int) -> bytes:
        """
        Returns the bytes of a share based on the given values of a share.
        """
        # Combine the version, threshold, and X-value into the same 2-byte
        # sequence.
        version_threshold_x = encoded_version + encoded_threshold + encoded_X
        version_threshold_x_bin = version_threshold_x.to_bytes(2, "big")

        # The share checksum is the first two bytes of the sha256 hash of all
        # previous bytes.
        share_bytes_before_checksum = [
            encoded_Y.to_bytes(32, "big"),
            seed_checksum,
            version_threshold_x_bin
        ]

        share_checksum = sha256(b"".join(share_bytes_before_checksum))[:2]
        share_checksum_int = int.from_bytes(share_checksum, "big")

        # Xor the share cheksum to the version, threshold, and X-value.
        version_threshold_x_xor = version_threshold_x ^ share_checksum_int

        # Join and return the share bytes.
        share_bytes = [
            encoded_Y.to_bytes(32, "big"),
            seed_checksum,
            version_threshold_x_xor.to_bytes(2, "big"),
            share_checksum
        ]

        return b"".join(share_bytes)


    @staticmethod
    def mnemonic_bytes(mnemonic_int: int) -> bytes:
        """
        Returns the bytes of a Mnemonic based on the given int representation.
        """
        return mnemonic_int.to_bytes(33, "big")


    @staticmethod
    def mnemonic_seed(mnemonic_int: int) -> bytes:
        """
        Returns the first 32 bytes of the Mnemonic based on the given int
        representation.
        """
        return mnemonic_int.to_bytes(33, "big")[:32]


    @staticmethod
    def mnemonic_checksum(mnemonic_int: int) -> bytes:
        """
        Returns the last bytes of a Mnemonic based on the given int
        representation.
        """
        return mnemonic_int.to_bytes(33, "big")[-1:]


    @staticmethod
    def mnemonic_hash(key_int: int) -> bytes:
        """
        Returns the full sha256 hash of a Mnemonic class instance.
        """
        key_bytes = key_int.to_bytes(32, "big")
        return sha256(key_bytes).digest()


from .bitcoinshamir import *
from .enums import Language, Checksum
from .exceptions import *
from .BIP39_List import BIP39_List
from .lagrange import Lagrange
from .decode import Decode
from .encode import Encode
from .mnemonic import Mnemonic
from .share import Share
from .point import Point
from .polynomial import Polynomial

wordlist = BIP39_List()


from .point import Point
from typing import List

class Lagrange:
    """
    Class containing static methods for caluclating lagrange interpolation.
    """
    @staticmethod
    def interpolate(points: List[Point], modulus: int, X: int) -> int:
        """
        Gets the Y-value using Lagrange interpolation according to the given
        list of points, over the finite field of the given modulus.
        """
        if not isinstance(modulus, int):
            raise TypeError("The given modulus was not of the int type.")

        if modulus < 3:
            raise ValueError("The given modulus was out of bounds.")

        if not isinstance(points, list):
            raise TypeError("The given points argument was not of type list.")

        if len(points) < 2:
            message = f"At least 2 points required. Only {len(points)} given."
            raise ValueError(message)

        for point in points:
            if not isinstance(point.X, int) or not isinstance(point.Y, int):
                raise TypeError("The given point values are not of type int.")
            elif point.X < 0 or point.Y < 0:
                raise ValueError(f"({point.X}, {point.Y}) has negative value.")

        # This will be the return value.
        cumulative_sum = 0

        # Sum loop.
        for point_a in points:
            cumulative_product = point_a.Y

            # Product loop.
            for point_b in [p for p in points if p is not point_a]:
                numerator = (X - point_b.X) % modulus
                denominator = (point_a.X - point_b.X) % modulus

                # Multiplicitive inverse using Fermat's little theorem.
                mul_inv = pow(denominator, modulus - 2, modulus)
                product = numerator * mul_inv % modulus
                cumulative_product = cumulative_product * product % modulus

            cumulative_sum = (cumulative_sum + cumulative_product) % modulus

        return cumulative_sum


from .enums import Checksum, Language

class ChecksumError(Exception):
    """
    Exception raised when a checksum is invalid.
    """
    def __init__(self, checksum_type: Checksum, first_checksum: bytes,
            second_checksum: bytes, *args: object) -> None:
        super().__init__(*args)
        self.checksum_type = checksum_type
        self.first_checksum = first_checksum
        self.second_checksum = second_checksum


    def __str__(self) -> str:
        message = "The {0} values did not match\n1:\t{1}\n2:\t{2}"

        return message.format(
            self.checksum_type.value, self.first_checksum, self.second_checksum
        )


class ThresholdError(Exception):
    """
    Exception raised when number of shares provided do not meet the threshold.
    """
    def __init__(self, threshold: int, actual: int, *args: object) -> None:
        super().__init__(*args)
        self.threshold = threshold
        self.actual = actual


    def __str__(self) -> str:
        message = "{0} does not meet the minimum threshold of {1}"
        return message.format(self.actual, self.threshold)


class LanguageError(Exception):
    """
    Exception raised when a word or language is not found.
    """
    def __init__(self, language: Language, *args: object) -> None:
        super().__init__(*args)
        self.language = language


    def __str__(self) -> str:
        return f"'{self.language}' not found in the current language list."


class WordlistError(Exception):
    """
    Exception raised when a word is not found in a BIP39 word list.
    """
    def __init__(self, word: str, language: Language, *args: object) -> None:
        super().__init__(*args)
        self.word = word
        self.language = language


    def __str__(self) -> str:
        return f"'{self.word}' not found in {self.language} word list."


from enum import Enum


class Language(str, Enum):
    ChineseSimplified = "chinese_simplified"
    ChineseTraditional = "chinese_traditional"
    Czech = "czech"
    English = "english"
    French = "french"
    Italian = "italian"
    Japanese = "japanese"
    Korean = "korean"
    Portuguese = "portuguese"
    Spanish = "spanish"


class Checksum(str, Enum):
    Mnemonic = "Mnemonic 8-bit Checksum"
    KeyValue = "Mnemonic 32-byte Checksum"
    ShareGroup = "Share Group 8-bit Checksum"
    Share = "Share Key 16-bit Checksum"


class Decode:
    @staticmethod
    def share_X(encoded_X: int) -> int:
        """
        Returns the actual X-value of a Share object based on the given
        encoded X-value.
        """
        # The decoded X-value starts at 2.
        return encoded_X + 2


    @staticmethod
    def share_threshold(encoded_threshold: int) -> int:
        """
        Returns the actual threshold of a Share object based on the given
        encoded threshold value.
        """
        # Get the encoded threshold value by truncating the last 7 bits. The
        # threshold is the second item encoded in a 2-byte sequence, followed by
        # the X-value.
        decoded_threshold = encoded_threshold >> 7

        # The decoded threshold starts at 2.
        return decoded_threshold + 2


    @staticmethod
    def share_version(encoded_version: int) -> int:
        """
        Returns the given version number as the encoded value in a Share
        object.
        """
        # Get the version number by truncating the last 11 bits. The version
        # number is the first item encoded in a 2-byte sequence, followed by the
        # threshold and the X-value.
        return encoded_version >> 11


    @staticmethod
    def mnemonic_int(seed: bytes, checksum: bytes) -> int:
        """
        Returns the integer representation of a Mnemonic class instance.
        """
        return int.from_bytes(seed + checksum, "big")


    @staticmethod
    def mnemonic_key(seed_or_full_mnemonic: bytes) -> int:
        """
        Returns the integer key of a Mnemonic class instance.
        """
        key_bytes = seed_or_full_mnemonic[:32]
        return int.from_bytes(key_bytes, "big")


class Polynomial:
    """
    Polynomial class for storing coefficients of a polynomial in a finite
    field.
    """
    def __init__(self, x0 = 0) -> None:
        """
        Creates a new Polynomial class, setting the x^0 coefficient to the given
        input. If no input is given, the x^0 coefficient is set to 0.
        """
        self.coefficients = [x0]


    def __repr__(self) -> str:
        """
        Returns a string representation of this Polynomial class instance.
        """
        components = []

        for i, coefficient in reversed(enumerate(self.coefficients)):
            if i > 0:
                components.append(f"{coefficient}x^{i}")
            else:
                components.append(f"{coefficient}")

        expression = " +\n".join(components)

        return f"Polynomial():\n{expression}"


    def solve(self, x: int, modulus: int) -> int:
        """
        Returns the Y value of the current polynomial coefficients based on the
        given X value.
        """
        result = 0

        for i, coefficient in enumerate(self.coefficients):
            result += coefficient * x ** i % modulus
        
        return result % modulus


from .exceptions import *
from .BIP39_List import BIP39_List
from .encode import Encode
from .decode import Decode
from .point import Point
from .enums import Language
from typing import List
from hashlib import sha256


PRIME_MODULUS = 2 ** 256 - 2 ** 32 - 977
wordlist = BIP39_List()
current_version = 0


class Share:
    """
    Share class for holding a single share of a Shamir (k of n) threshold
    scheme. Each share is a coordinate, with an (x, y) value, where x is an
    integer, greater than 1. The secret key is the y value at x = 0. A checksum
    of the key is the value at x = 1. Share values start at x = 2.

    One share is 37 bytes, represented as 27 words. Each word is 11 bits. The
    297th bit from the last word is truncated. The first 32 bytes encode the
    y value of the share. The next byte is the original seed's checksum. The
    next 2 bytes hold  the version, threshold, and x value of the share. The
    version is 5 bits, starting with a value of 0; the threshold is 4 bits,
    starting at 2; and the x value is 7 bits, starting at 2. The last 2 bytes
    are a checksum of the share. Finally, the share checksum is applied as an
    xor to the prior 2-bytes containing the version, threshold, and x value in
    order to avoid long sequences of zeros.

    296-bit share structure:
    [256: Y value]
    [8: Seed checksum]
    [5: Version xor checksum]
    [4: Threshold xor checksum]
    [7: X value xor checksum]
    [16: Checksum]
    """
    def __init__(
            self, point: Point, threshold: int, seed_checksum: bytes,
            version: int = current_version) -> None:
        """
        Initializes an instance of the Share class, and assignes values for the
        (x, y) coordinate, as well as the version, threshold, and checksums. The
        threshold must be between [2, 17]. The X value must be between [2, 257].
        """
        if not isinstance(threshold, int):
            raise TypeError("The given threshold argument is not of type int.")

        if threshold < 2 or threshold > 17:
            raise ValueError("The given index argument is out of bounds.")

        if not isinstance(point, int):
            raise TypeError("The given point argument is not of type Point.")

        if not isinstance(point.X, int):
            raise TypeError("The given X argument is not of type int.")

        if point.X < 2 or point.X > 257:
            raise ValueError("The given X argument is out of bounds.")

        if not isinstance(point.Y, int):
            raise TypeError("The given Y argument is not of type int.")

        if point.Y < 1 or point.Y >= PRIME_MODULUS:
            raise ValueError("The given Y argument is out of bounds.")
        
        if len(seed_checksum) < 1:
            raise ValueError("The given seed_checksum was not at least 1 byte.")

        # Encode the 2-byte sequence containing the version, threshold, and
        # X-value in order to get the share's checksum.
        encoded_version = Encode.share_version(version)
        encoded_threshold = Encode.share_threshold(threshold)
        encoded_x = Encode.share_X(point.X)
        
        # Get the share checksum, which is the last two bytes of the fully
        # encoded share.
        share_bytes = Encode.share_bytes(
                encoded_x, point.Y, encoded_threshold, seed_checksum[:1],
                encoded_version
            )

        share_checksum = share_bytes[-2:]

        # Assign instance variables.
        self.seed_checksum = seed_checksum[:1]
        self.point = point
        self.threshold = threshold
        self.version = version
        self.share_checksum = share_checksum
    

    @classmethod
    def from_share_phrase(
            cls, phrase: List[str], language: Language
            ) -> "Share":
        """
        Returns an instance of a Share class according to the given share
        phrase. Raises an error if the language is not in the current language
        list, the mnemonic phrase has invalid words, or has the wrong number of
        words.
        """
        if not isinstance(language, Language):
            raise ValueError(f"{language} is not in the language list.")

        if not isinstance(phrase, list):
            raise TypeError("The given phrase was not of the list[str] type.")
        
        if len(phrase) != 27:
            raise ValueError("The given share phrase did not have 27 words.")

        for word in phrase:
            if language not in wordlist.get_language(word):
                raise WordlistError(word, language)

        indices = [wordlist.get_word_index(word, language) for word in phrase]
        
        # Add words from left to right, shifting the added words to the left by
        # 11 bits each iteration.
        share_int = 0

        for word_index in indices:
            share_int <<= 11
            share_int += word_index
        
        # Remove the last bit. A 27-word phrase has an extra bit that spills
        # over the 37 bytes that should be ignored.
        share_int >>= 1

        # Truncate the right 40 bits to get the Y-value of the share. The
        # Y-value is the left-most 256 bits, with 40 bits to its right.
        y_int = share_int >> 40

        # The seed checksum value is the 5th byte from the right.
        seed_checksum_int = (share_int & 0xFF_00_00_00_00) >> 32
        seed_checksum_bin = seed_checksum_int.to_bytes(1, "big")

        # The next 2 bytes are version, threshold, and X-value.
        version_threshold_x_val = (share_int & 0xFF_FF_00_00) >> 16

        # The final 2 bytes are the share checksum.
        share_checksum_int = share_int & 0xFF_FF

        # The version, threshold, and X-value are encoded with an xor of the
        # share's checksum to avoid excessive zeros in the share. This xor needs
        # to be removed before decoding each value.
        version_threshold_x_val ^= share_checksum_int

        version_encoded = version_threshold_x_val & 0b11111_0000_0000000
        threshold_encoded = version_threshold_x_val & 0b00000_1111_0000000
        x_val_encoded = version_threshold_x_val & 0b00000_0000_1111111

        version_int = Decode.share_version(version_encoded)
        threshold_int = Decode.share_threshold(threshold_encoded)
        x_int = Decode.share_X(x_val_encoded)

        point = Point(x_int, y_int)

        return cls(point, threshold_int, seed_checksum_bin, version_int)


    def to_bytes(self) -> bytes:
        """
        returns the 37-byte representation of the current Share instance.
        """
        # The version, threshold, and X-value are encoded in the same 2-byte
        # sequence.
        encoded_version = Encode.share_version(self.version)
        encoded_threshold = Encode.share_threshold(self.threshold)
        encoded_x = Encode.share_X(self.point.X)

        share_bytes = Encode.share_bytes(
                encoded_x, self.point.Y, encoded_threshold, self.seed_checksum,
                encoded_version
            )
        
        return share_bytes

    
    def get_word(self, index: int, language: Language) -> str:
        """
        Returns the word at the given zero-based index of this Share class
        instance. Raises error if the index is not between 0 and 26.
        """
        if not isinstance(index, int):
            raise TypeError("The index argument given is not an int.")

        if index < 0 or index > 26:
            raise IndexError("The index argument given is out of bounds.")
            
        if not isinstance(language, Language):
            raise ValueError(f"{language} is not in the language list.")

        # The version, threshold, and X-value are encoded in the same 2-byte
        # sequence.
        version_int = Encode.share_version(self.version)
        threshold_int = Encode.share_threshold(self.threshold)
        x_int = Encode.share_X(self.X)

        share_bytes = Encode.share_bytes(
                x_int, self.Y, threshold_int, self.seed_checksum, version_int
            )
        
        # Rehash the share to get the extra bit for the 27th word.
        version_threshold_x_int = version_int + threshold_int + x_int
        version_threshold_x_bin = version_threshold_x_int.to_bytes(2, "big")

        share_bytes_before_checksum = [
            self.Y.to_bytes(32, "big"),
            self.seed_checksum,
            version_threshold_x_bin
        ]

        third_checksum_byte = sha256(b"".join(share_bytes_before_checksum))[2:3]

        # Join the share bytes and remove 7 bits from the third byte of the
        # share checksum.
        share_byte_array = [
            share_bytes,
            third_checksum_byte
        ]

        share_bin = b"".join(share_byte_array)
        share_int = int.from_bytes(share_bin, "big")
        share_int >>= 7

        # Determine number of bits to truncate from the right based the word's
        # position in the phrase.
        max_words = 27
        word_position = index + 1
        remove_word_count = max_words - word_position
        truncated_share_int = share_int >> (remove_word_count * 11)

        # Get word index and return word text.
        word_bitmask = 0b1111_1111_111
        word_index = truncated_share_int & word_bitmask
        word = wordlist.get_word(word_index, language)

        return word


    # def copy(self) -> "Share":
    #     """
    #     Returns a copy of the Share object.
    #     """
    #     return Share(self.point.copy(), self.threshold, self.seed_checksum, self.version)


    def is_valid(self) -> bool:
        pass


class Point:
    """
    Point class for storing X, Y coordinates.
    """
    def __init__(self, X: int = 0, Y: int = 0) -> None:
        """
        Creates a new Point class, setting the X & Y values both to 0.
        """
        self.X = X
        self.Y = Y


    def __repr__(self) -> str:
        """
        Returns a string representation of this Point class instance.
        """
        return f"Point({self.X}, {self.Y})"


    # def copy(self):
    #     """
    #     Returns a copy of the Point object.
    #     """
    #     return Point(self.X, self.Y)